C	Test_Shotgun.FOR
C	8-MAY-1996 David Mathog, Division of Biology, Caltech
C
C	Copyright 1996, David Mathog
C	This program may be used for research purposes and may be freely 
C	copied, but it may not be incorporated 	in part or in whole in any
C	commercial product. 
C
C	This program is a platform for testing various shotgun
C	sequencing strategies.  It is also a cost estimation tool.
C	Various types of clones may be added in phases, and then
C	the program will complete the project to the point where
C	everything has been sequenced at least once in each direction.
C	The internal model is a 1Mb sequence, which is scaled
C	up for sequences bigger than that.  For smaller sequences,
C	a subsequence is used.  Note that scaling can produce some
C	quantization effects - if you ask for 1 clone in the 3.5Gb
C	human genome it will scale that to 3500 - it cannot enter
C	less than one clone per phase, and 3.5Gb/3500 = 1 Mb.
C
C	Most terms have common meanings, but "MESH" does not.  It
C	refers to the mesh of overlapping clones.  This is not quite the
C	same as a contig, as a MESH includeds overlaps like: A B A B,
C	where A,B are the sequenced ends.  The mesh can be converted to a
C	contig by filling in either one of the sequences.  
C
C	Compilation instructions.
C
C	OpenVMS:
C	    $ for/extend/nolis test_shotgun
C	    $ link/nomap       test_shotgun
C
C	Irix, Digital Unix (probably other Unix's as well):
C	    % mv test_shotgun.for test_shotgun.f
C	    % f77 -o test_shotgun -extend_source test_shotgun.f
C
C	SunOS 4.1
C	    % f77 -o test_shotgun -e -xl test_shotgun.f
C	    (During compilation it will complain about "carriagecontrol"
C	     and during a run it will issue one warning:
C             Warning: open: record length ignored on sequential access
C            Ignore these, as the results look ok.)
C
C       Others:  It uses some F77 extensions, notably long program
C	         lines, tab spacing.  Most F77s support these.
C
C	Here is a test script to feed it: (first strip the leading CC!)
CCrun test_shotgun     (or start it as appropriate for OS)
CCkillme.dat
CCThis is a do_test result
CC3500000000
CC15.00
CC500,15.0
CC4000,5.00
CC50
CC2000,0.10
CC500,15.0
CC7000000
CC0
C
	implicit none
	integer MAXSEQ,MAXCLONE
	parameter (MAXCLONE = 100000)
	parameter (MAXSEQ = 1000000)
	character*256 outfile
	character*60 projstring
	structure /triad/
	   integer*4 begin
	   integer*4 end
	   integer*4 id
	end structure
	integer*4 report(3,maxseq)
	record /triad/ frags(maxclone)
	record /triad/ clones(maxclone)
	record /triad/ scratch(maxclone)
	integer*4 mesh(maxclone)
	integer*4 olen
	integer*4 phase             !Phase in sequencing project
	integer*4 clen              !shotgun clone length for current phase
	integer*4 eclen             !clone length for clones selected  to fill
	                            !between meshes
	double precision  ccost     !cost to obtain and purify each shotgun clone
	double precision  sumccost  !sum of "
	double precision  ecost     !cost to obtain and purify each interMesh clone
	double precision  sumecost  !sum of "
	integer*4         slen      !sequence length (from ends of shotgun clone)
	double precision  scost     !cost to sequence EACH end of a shotgun clone
	double precision  sumscost  !sum of "
	integer*4         ilen      !internal sequence length (from a primer)
	integer*4         islen     !internal sequence step length (from a primer)
	                            !is sequence length minus required overlap
	double precision  pcost     !cost to make an internal primer
	double precision  sumpcost  !sum of "
	double precision  icost     !cost to sequence from a custom primer
	double precision  sumicost  !sum of "
	integer*4         cnum      !number of clones to do
	integer*4         ecnum     !external value for number of clones to do
	integer*4         overlap   !min. overlap to merge a Mesh/Contig
	integer*4         userlength!length of DNA in user's project
	double precision  euserlength  !just for reading length from the user
	double precision  uscale    !convert to user supplied project length
	integer*4 fragpointer,clonepointer
	integer*4 ID    !counter for fragment ID's
	integer*4 pos,pos2,minpos,maxpos,span,first,back,test,temp
	real*4    scale
C	some compilers need RAN declared as a real, others don't care
	real*4    ran
	integer*4 i,j
	integer*4 again
	integer*4 covered,included
C
C	Variables for:
C	e* = number of external (between MESH) clones needed to close
C	o* = gaps BETWEEN Mesh
C	i* = gaps WITHIN Mesh
C	s* = single direction (!) sequences
C	b* = overlapping sequences (in opposite directions)
C	v* = overlapping sequences (in same direction)
C	m* = meshes
C	k* = contigs
C
	integer*4 enum
	integer*4 osum,ofirst,omax,onum,osteps
	integer*4 isum,ifirst,imax,inum,isteps
	integer*4 bsum,bfirst,bmax,bnum,bsteps
	integer*4 ssum,sfirst,smax,snum,ssteps
	integer*4 vsum,vfirst,vmax,vnum
	logical   oingap,iingap,bingap,singap,vingap
	logical   paint
	integer*4 msum,mmax,mnum,msumb,mmaxb
	integer*4 ksum,kmax,knum,ksumb,kmaxb
c
c	needed for Sun compiler
c
	real*4    secnds
c
        temp = 100*secnds(0.0) + 1

c
c
	write(6,*)'Test_Shotgun'
	write(6,*)'  This program tests various cloning strategies on a'
	write(6,*)'  project.  Internally, it models this project with an area'
	write(6,*)'  of ',maxseq,' bases of DNA.  If your project is larger' 
	write(6,*)'  than this the results will be scaled to match.  If your'
	write(6,*)'  project is smaller then only part of this region will'
	write(6,*)'  be utilized.'
	write(6,*)'  '
	write(6,*)'  The total number of fragments (after scaling) that'
	write(6,*)'  can be modeled is',maxclone/2,',(which corresponds to'
	write(6,*)'  a maximum density of one clone start every 40 base pairs.'
	write(6,*)'  '
	write(6,*)'  This program uses the term Mesh in addition to Contig. Usually' 
	write(6,*)'  Contigs refer to overlapping stretches of sequence.'
	write(6,*)'  During the simulation the program can tell which'
	write(6,*)'  clones overlap even if there is not enough common'
	write(6,*)'  sequence to form a classical Contig.  In order to avoid'
	write(6,*)'  confusion, each such region of overlapping clones'
	write(6,*)'  (not usually the same as overlapping sequences) is called'
	write(6,*)'  a Mesh.'
	write(6,*)'  '
	write(6,*)'  You may vary at each phase:'
	write(6,*)'      the number of clones added'
	write(6,*)'      the size (unit cost) of these clones'
	write(6,*)'      the number of bases (cost) for each end sequence'
	write(6,*)'  '
	write(6,*)'  After all phases have been completed the program will'
	write(6,*)'  estimate the costs and steps needed to finish the project.'
	write(6,*)'  The project is finished when every region has been '
	write(6,*)'  sequenced at least once in each direction and the minimum'
	write(6,*)'  end overlaps are satisfied.'
	write(6,*)'  '
	write(6,*)'  Note that Mesh/Contig numbers only scale if the number'
	write(6,*)'  is more than one, and the scale value is greater than 1'
	write(6,*)'  That is, multiple Mesh/Contigs are found and the project is'
	write(6,*)'  bigger than the internal work area.'
	write(6,*)'  '
	write(6,*)'  You may optionally specify costs:'
	write(6,*)'    per clone         (isolation, purification, storage)'
	write(6,*)'    per sequenced end (sequencing, data entry)'
	write(6,*)'  '
	write(6,*)'  In those locations where costs are entered, the prompt'
	write(6,*)'    will include (,cost) and you would respond'
	write(6,*)'    with (being sure to include the decimal point):  '
	write(6,*)'       400, 12.35'
	write(6,*)'    for "400 bases at a cost of 12.35"'
	write(6,*)'    If no cost is provided it defaults initially to 0.0'
	write(6,*)'    and subsequently to the previous assigned value'
	write(6,*)'  '
c
	write(6,*)'Input the name of the output file'
	read(5,'(q,a)')olen,outfile(1:olen)
c
	open(unit=12,file=outfile(1:olen)
	1 ,form='FORMATTED',carriagecontrol='LIST',status='NEW'
	2 ,recl = 132)
c
	icost    = 0.0
	pcost    = 0.0
	scost    = 0.0
	ccost    = 0.0
	sumicost = 0.0
	sumpcost = 0.0
	sumscost = 0.0
	sumccost = 0.0
	write(6,*)' '
	write(6,*)
	1 'Input one text line that describes this project (<=60 characters)'
	read(5,2004)projstring
	write(6,*)' '
	write(6,*)'Input the amount of project DNA, in base pairs'
	read(5,*)euserlength
	if(euserlength .lt. 0.0)then
	  stop 'Impossible length for DNA!!!'
	else if(euserlength .gt. maxseq)then
	  uscale = euserlength/float(maxseq)
	  userlength = float(maxseq)
	else
	  userlength = euserlength
	  uscale = 1.0
	end if

2000	format(i9,f9.3)
2002	format(f9.3)
2004	format(a)
	write(6,*)' '
	write(6,*)'The next set of questions apply to finishing off a '
	write(6,*)'project once all initial phases have been completed.'
	write(6,*)' '
	write(6,*)'Input the cost of obtaining an oligonucleotide primer'
	read(5,2002)pcost
	write(6,*)' '
	write(6,*)'Input the amount of sequence recoverable from a'
	write(6,*)'  primer walking sequencing step (,cost)'
	read(5,2000)ilen,icost
	write(6,*)' '
	write(6,*)'Input the size of clones that will be used to close'
	write(6,*)'  gaps between Mesh/Contigs (,cost).  (NOTE.  if clones'
	write(6,*)'  are to be selected by hybridization with a single'
	write(6,*)'  probe then this size should be 1/2 the actual size'
	write(6,*)'  of the clones as each clone will be randomly positioned'
	write(6,*)'  with respect to that probe so, on average, only 1/2 of'
	write(6,*)'  each clone will be new sequence.)'
	read(5,2000)eclen,ecost
	write(6,*)' '
	write(6,*)'Input the minimum overlap to merge a Mesh/Contig'
	write(6,*)'AND for overlap in primer walking'
	read(5,*)overlap
	write(6,*)' '
	islen = ilen - overlap
	if(islen .le. 0)stop 
	1 'Fatal error - overlap and internal step size incompatible'
c
	do i = 1,userlength
	  report(1,i) = 0     !forward coverage
	  report(2,i) = 0     !reverse coverage
	  report(3,i) = 0     !inclusion in a clone - no sequence
	end do
	do i = 1,maxclone
	  mesh(i) = 0
	end do
c
	write(12,1005)' '
	write(12,1005)' Project: '//projstring
	write(12,1005)' '
	write(12,1005)' Project constants'
	write(12,1005)' '
	euserlength = float(userlength)*uscale
	write(12,1004)'   Genome size (base pairs):',euserlength
	write(12,1002)'   Scaling from internal subsequence:',uscale
	write(12,1010)'   Required overlap for Mesh merge:',overlap
	write(12,1010)'   Bases of sequence from a custom primer:',islen
	write(12,1010)'   Bases in clones used to close mesh gaps:',eclen
	write(12,1002)'   Cost to obtain a clone outside a Mesh/contig:',ecost
	write(12,1002)'   Cost to make a custom primer:',pcost
	write(12,1002)'   Cost to sequence from a custom primer:',icost
	write(12,1005)' '
c
	phase = 0
	paint  = .true.
	fragpointer = 0
	clonepointer = 0
	ID = 0
c
	do while(paint)
	phase = phase + 1
	write(6,*)' '
	write(6,*)'DNA sequencing project phase: ',phase
	write(6,*)' '
	write(6,*)'Input the size of each clone (,cost per clone)'
	read(5,2000)clen,ccost
	write(6,*)'Input the amount of sequence from each end(,cost)'
	read(5,2000)slen,scost
	if(slen .gt. clen)stop 'Fatal error, more sequence than clone!!!'
	write(6,*)'Input the number of such clones to process'
	read(5,*)ecnum
	cnum = int(ecnum/uscale)
	if(cnum .le. 0)then
	  write(6,*)'TEST_SHOTGUN: warning'
	  write(6,*)'  scaled # of clones was zero, using one instead'
	  cnum = 1
	  ecnum = float(cnum) * uscale
	end if
	minpos = 1 - clen
	maxpos = userlength + clen
	span = 	maxpos - minpos
	write(12,1005)' '
	write(12,1010)' Phase :',phase
	write(12,1010)'   Number of clones added:',ecnum
	write(12,1010)'   Size of each clone added:',clen
	write(12,1010)'   Bases of sequence from each end primer:',slen
	write(12,1002)'   Number of genome units of coverage:'
	1   ,ecnum * dfloat(clen)/euserlength
	write(12,1002)'   Number of genome units of sequence:'
	1   ,ecnum * dfloat(slen) * 2.0 /euserlength
	write(12,1010)' '
	write(12,1002)'   Unit cost: clone isolation and preparation',ccost
	write(12,1002)'   Unit cost: end primer sequencing:',scost
	write(12,1010)' '
	write(12,1002)'   Total phase cost for clone preparation:',ccost*ecnum
	write(12,1002)'   Total phase cost for end sequencing:',scost*2.0*ecnum
c
c
	do i = ID + 1 ,ID + cnum
	   sumccost = sumccost + ccost
	   sumscost = sumscost + scost + scost
	   pos = minpos + span*ran(temp)
	   pos2 = pos + clen - slen
	   back = pos + clen - 1
c
c	paint the "forward" region
c
	   do j = pos,pos + slen - 1
	     if(j .ge. 1 .and. j .le. userlength)report(1,j) = report(1,j) + 1
	   end do
c
c	paint the "reverse" region
c
	   do j = pos2,pos2 + slen - 1
	     if(j .ge. 1 .and. j .le. userlength)report(2,j) = report(2,j) + 1
	   end do
c
c	paint the "included" region on the total DNA
c
	   do j = pos,pos + clen - 1
	     if(j .ge. 1 .and. j .le. userlength)report(3,j) = report(3,j) + 1
	   end do
c
c	store the clone
c
	clonepointer = clonepointer + 1
	if(clonepointer .le. maxclone)then
	  clones(clonepointer).begin  = pos
	  clones(clonepointer).end    = back
	  clones(clonepointer).id     = i
	else
	  stop 'TEST_SHOTGUN: fatal error, too many clones'
	end if
c
c	store the two sequence fragments from this piece of DNA if they
c	are bigger than overlap
c
	   if(pos + slen - 1 .ge. overlap .and. 
	1     pos .le. userlength - overlap - 1)then
	     fragpointer = fragpointer + 1
	     if(fragpointer .ge. maxclone)then
	       stop 'TEST_SHOTGUN: fatal error, too many sequences'
	     else
	       frags(fragpointer).begin = max(1,pos)
	       frags(fragpointer).end   = min(userlength,pos + slen - 1)
	       frags(fragpointer).id    = i
	     end if
	   else  !front piece truncated too small, pretend it isn't there
	     pos = 0
	   end if
	   if(pos2 + slen - 1 .ge. overlap .and. 
	1    pos2 .le. userlength - overlap - 1)then
	     fragpointer = fragpointer + 1
	     if(fragpointer .ge. maxclone)then
	       stop 'TEST_SHOTGUN: fatal error, too many sequences'
	     else
	       frags(fragpointer).begin = max(1,pos2)
	       frags(fragpointer).end = min(userlength,pos2 + slen - 1)
	       if(pos .eq. pos2 .or. pos .eq. 0)then
	         frags(fragpointer).id = i
	       else
	         frags(fragpointer).id  = i + maxclone
	       end if
	     end if
	   end if
	end do
	ID = ID + CNUM
	write(6,*)'0 exits, any other number to add more fragments'
	read(5,*)again
	if(again .eq. 0)paint = .false.
	end do !paint
c
c
c	add everything up
c
	covered  = 0
	included = 0
	enum   = 0
	osum   = 0
	onum   = 0
	omax   = 0
	oingap    = .false.
	osteps    = 0
	isum   = 0
	inum   = 0
	imax   = 0
	iingap    = .false.
	isteps    = 0
	ssum   = 0
	snum   = 0
	smax   = 0
	singap    = .false.
	ssteps    = 0
	bsum   = 0
	bnum   = 0
	bmax   = 0
	bingap    = .false.
	bsteps    = 0
	vsum   = 0
	vnum   = 0
	vmax   = 0
	msum   = 0
	mmax   = 0
	msumb   = 0
	mmaxb   = 0
	mnum   = 0
	ksum   = 0
	knum   = 0
	knum   = 0
	kmax   = 0
	ksumb  = 0
	kmaxb  = 0
	vingap    = .false.
	do i = 1,userlength
c
c	There are several possible states this position might be in,
c	operate as appropriate in each case on the o*,i*,s*,b*,v* variables
c
c
c	first case, has forward sequence, maybe reverse sequence too
c
	  if(report(1,i) .gt. 0)then
	     covered  = covered + 1
	     included = included + 1
c
	     if(oingap)then
	       oingap = .false.
	       if(i - ofirst .gt. omax)omax = i - ofirst
	       osteps = osteps + (i - ofirst - 1)/islen + 1
	       enum = enum + 1 + (i - ofirst - 1)/eclen
	     end if
c
	     if(iingap)then
	       iingap = .false.
	       if(i - ifirst .gt. imax)imax = i - ifirst
	       isteps = isteps + (i - ifirst - 1)/islen + 1
	     end if
c
	     if(report(2,i) .gt. 0)then
	         if(singap)then
	           singap = .false.
	           if(i - sfirst .gt. smax)smax = i - sfirst
	           ssteps = ssteps + (i - sfirst - 1)/islen + 1
	         end if
c
	         bsum = bsum + 1
	         if(.not. bingap)then
	           bingap = .true.
	           bnum = bnum + 1
	           bfirst = i
	         end if
c
	         if(vingap)then
	           vingap = .false.
	           if(i - vfirst .gt. vmax)vmax = i - vfirst
	         end if
c
	     else
c
	         if(report(1,i) .eq. 1)then
	           ssum = ssum + 1
	           if(.not. singap)then
	             snum = snum + 1
	             singap = .true.
	             sfirst = i
	           end if
	         else
	           if(singap)then
	             singap = .false.
	             if(i - sfirst .gt. smax)smax = i - sfirst
	             ssteps = ssteps + (i - sfirst - 1)/islen + 1
	           end if
	         end if
c
	         if(bingap)then
	           bingap = .false.
	           if(i - bfirst .gt. bmax)bmax = i - bfirst
	         end if
c
	         if(report(1,i) .ge. 2)then
	           vsum = vsum + 1
	           if(.not. vingap)then
	             vingap = .true.
	             vnum = vnum + 1
	             vfirst = i
	           end if
	         else
	           if(vingap)then
	             vingap = .false.
	             if(i - vfirst .gt. vmax)vmax = i - vfirst
	           end if
	         end if
c
	     end if
c
c	second case, has reverse sequence , but NO forward sequence
c	
	  else if(report(2,i) .gt. 0)then
	     covered  = covered + 1
	     included = included + 1
c
	     if(oingap)then
	       oingap = .false.
	       if(i - ofirst .gt. omax)omax = i - ofirst
	       osteps = osteps + (i - ofirst - 1)/islen + 1
	       enum = enum + 1 + (i - ofirst - 1)/eclen
	     end if
c
	     if(iingap)then
	       iingap = .false.
	       if(i - ifirst .gt. imax)imax = i - ifirst
	       isteps = isteps + (i - ifirst - 1)/islen + 1
	     end if
c
	     if(report(2,i) .eq. 1)then
	       ssum = ssum + 1
	       if(.not. singap)then
	         snum = snum + 1
	         singap = .true.
	         sfirst = i
	       end if
	     else
	       if(singap)then
	         singap = .false.
	         if(i - sfirst .gt. smax)smax = i - sfirst
	         ssteps = ssteps + (i - sfirst - 1)/islen + 1
	       end if
	     end if
c
	     if(bingap)then
	       bingap = .false.
	       if(i - bfirst .gt. bmax)bmax = i - bfirst
	     end if
c
	     if(report(2,i) .ge. 2)then
	       vsum = vsum + 1
	       if(.not. vingap)then
	         vingap = .true.
	         vnum = vnum + 1
	         vfirst = i
	       end if
	     else
	       if(vingap)then
	         vingap = .false.
	         if(i - vfirst .gt. vmax)vmax = i - vfirst
	       end if
	     end if
c
c	Third case, it hasn't been sequenced yet (no forward, no reverse
C	sequence), but it IS inside a clone
c
	  else if(report(3,i) .gt. 0)then
c
	     isum = isum + 1
	     included = included + 1
c
	     if(oingap)then
	       oingap = .false.
	       if(i - ofirst .gt. omax)omax = i - ofirst
	       osteps = osteps + (i - ofirst - 1)/islen + 1
	       enum = enum + 1 + (i - ofirst - 1)/eclen
	     end if
c
	     if(.not. iingap)then
	       inum = inum + 1
	       iingap = .true.
	       ifirst = i
	     end if
c
	     if(singap)then
	       singap = .false.
	       if(i - sfirst .gt. smax)smax = i - sfirst
	       ssteps = ssteps + (i - sfirst - 1)/islen + 1
	     end if
c
	     if(bingap)then
	       bingap = .false.
	       if(i - bfirst .gt. bmax)bmax = i - bfirst
	     end if
c
	     if(vingap)then
	       vingap = .false.
	       if(i - vfirst .gt. vmax)vmax = i - vfirst
	     end if
c
c	last case.  This region is not included inside any clone and so
C	has not been sequenced
c
	  else
c
	     osum = osum + 1
	     if(.not. oingap)then
	       onum = onum + 1
	       oingap = .true.
	       ofirst = i
	     end if
c
	     if(iingap)then
	       iingap = .false.
	       if(i - ifirst .gt. imax)imax = i - ifirst
	       isteps = isteps + (i - ifirst - 1)/islen + 1
	     end if
c
	     if(singap)then
	       singap = .false.
	       if(i - sfirst .gt. smax)smax = i - sfirst
	       ssteps = ssteps + (i - sfirst - 1)/islen + 1
	     end if
c
	     if(bingap)then
	       bingap = .false.
	       if(i - bfirst .gt. bmax)bmax = i - bfirst
	     end if
c
	     if(vingap)then
	       vingap = .false.
	       if(i - vfirst .gt. vmax)vmax = i - vfirst
	     end if
	  end if
c
	end do
c
c	handle any that are open at the end of the sequence
c
	i = userlength
c
	if(oingap)then
	  oingap = .false.
	  if(i - ofirst .gt. omax)omax = i - ofirst
	  osteps = osteps + (i - ofirst - 1)/islen + 1
	  enum = enum + 1 + (i - ofirst - 1)/eclen
	end if
c
	if(iingap)then
	  iingap = .false.
	  if(i - ifirst .gt. imax)imax = i - ifirst
	  isteps = isteps + (i - ifirst - 1)/islen + 1
	end if
c
	if(singap)then
	  singap = .false.
	  if(i - sfirst .gt. smax)smax = i - sfirst
	  ssteps = ssteps + (i - sfirst - 1)/islen + 1
	end if
c
	if(bingap)then
	  bingap = .false.
	  if(i - bfirst .gt. bmax)bmax = i - bfirst
	end if
c
	if(vingap)then
	  vingap = .false.
	  if(i - vfirst .gt. vmax)vmax = i - vfirst
	end if

c	**************************************************
C
c	This section figures out the number of contigs
c
c	sort the fragments by their beginning positions.  Then by their end
C	positions if necessary
c
	call binsort(frags,scratch,fragpointer,0)
c
	j = 1
	do while(j .le. fragpointer)
	  i = j
	  do while(frags(j).begin .eq. frags(i+1).begin .and.
	1          i .lt. fragpointer)
	     i = i + 1
	  end do
	  if(i - j .ge. 1)call binsort(frags(j),scratch,i-j+1,1)
	  j = i + 1
	end do
c
c	at this point all fragments are in order, so count the contigs
c
c
	first = 1
	back  = frags(1).end
	do i = 2, fragpointer
	  if(back - frags(i).begin + 1 .ge. overlap)then
	     back = max(back,frags(i).end)
	  else
	     ksum  = ksum + i - first
	     kmax  = max(kmax, i - first)
	     ksumb = ksumb + back - frags(first).begin + 1
	     kmaxb = max(kmaxb, back - frags(first).begin + 1)
	     knum = knum + 1
	     first = i
	     back  = frags(i).end	     
	  end if
	end do
c
c	there will always be one left dangling, so finish it off
c
	ksum  = ksum + fragpointer - first + 1
	kmax  = max(kmax, fragpointer - first + 1)
	ksumb = ksumb + back - frags(first).begin + 1
	kmaxb = max(kmaxb, back - frags(first).begin + 1)
	knum = knum + 1
c
c	**************************************************
c	This sectin figures out the number of MESHes.  A mesh
c	is a contig made up of clone coverage, not sequence coverage.
c	We can calculate what the MESH value is, but in a real
c	experiment this usually cannot be determined from the data at hand.
c
c	sort the fragments by their beginning positions.  Then by their end
C	positions if necessary
c
	call binsort(clones,scratch,clonepointer,0)
c
	j = 1
	do while(j .le. clonepointer)
	  i = j
	  do while(clones(j).begin .eq. clones(i+1).begin .and.
	1          i .lt. clonepointer)
	     i = i + 1
	  end do
	  if(i - j .ge. 1)call binsort(clones(j),scratch,i-j+1,1)
	  j = i + 1
	end do
c
c	at this point all clones are in order, so count the contigs
c
c
	first = 1
	back  = clones(1).end
	do i = 2, clonepointer
	  test = back - clones(i).begin + 1
	  if(test .ge. overlap)then
	     back = max(back,clones(i).end)
	  else
	     msum = msum + i - first
	     mmax = max(mmax, i - first)
	     msumb = msumb + back - clones(first).begin + 1
	     mmaxb = max(mmaxb, back - clones(first).begin + 1)
	     mnum = mnum + 1
c
c	If these overlap, or are immediately adjacent, they won't show
c	up as intermesh gaps above, so fix that here.  Have to add one
c	more clone (enum), one more intermesh gap (onum), and one more
c	sequence (ostep) to merge these two meshes
c
	     if(test .ge. 0)then
	        onum = onum + 1
	        enum = enum + 1
	        osteps = osteps + 1
	     end if
	     first = i
	     back  = clones(i).end	     
	  end if
	end do
c
c	there will always be one left dangling, so finish it off
c
	msum = msum + clonepointer - first + 1
	mmax = max(mmax, clonepointer - first + 1)
	msumb = msumb + back - clones(first).begin + 1
	mmaxb = max(mmaxb, back - clones(first).begin + 1)
	mnum = mnum + 1


c
c	**************************************************
c
c	This section reports the results
c
1000	format(a,t54,f9.7)
1002	format(a,t50,f13.2)
1004	format(a,t50,f14.0)
1005	format(a)
1010	format(a,t50,i13)
	scale = 1.0/float(userlength)
	write(12,1005)' '
	write(12,1010)' Project status after phase:',phase
	write(12,1005)' '
	write(12,1005)' Summary of coverage fractions:'
	write(12,1000)'   included in clones:',float(included)*scale
	write(12,1000)'   sequenced:',float(covered)*scale
	write(12,1000)'     both directions:'
	1  ,float(bsum)*scale
	write(12,1000)'     multiply, one direction:',float(vsum)*scale
	write(12,1000)'     once:',float(ssum)*scale
	write(12,1000)'   gaps:',float(isum+osum)*scale
	write(12,1000)'     IntraMesh:',float(isum)*scale
	write(12,1000)'     InterMesh:',float(osum)*scale
	write(12,1005)' '
	write(12,1005)' Mesh information'
c
c	If it reduced to a single Mesh then rescale MAX and average
c	to external lengths/number of fragments and report a single mesh
c	containing the number of clones scaled, and of length equal to
c	total genome length.
c	However, if there were more than one Meshes then scale the
c	number of Meshes to match the external lengths and fragments
c	but DO NOT scale the max size and average size, which should be
c	nearly correct.
c
	if(mnum .gt. 1)then
	  write(12,1010)'   number (Meshes):',int(mnum*uscale)
	  write(12,1010)'   max size (clones):',mmax
	  write(12,1010)'   avg. size (clones):',int(float(msum)/float(mnum))
	  write(12,1010)'   max size (bp):',mmaxb
	  write(12,1010)'   avg. size (bp):',int(float(msumb)/float(mnum))
	else
	  write(12,1010)'   number (Meshes):',1
	  write(12,1010)'   max size (clones):',int(mmax*uscale)
	  write(12,1010)'   avg. size (clones):',int(msum*uscale)
	  write(12,1004)'   max size (bp):',euserlength
	  write(12,1004)'   avg. size (bp):',euserlength
	end if
	write(12,1005)' '
	write(12,1005)' Contig information'
c
c
c	contigs are treated like meshes, see above
c
	if(knum .gt. 1)then
	  write(12,1010)'   number (Contigs):',int(knum*uscale)
	  write(12,1010)'   max size (sequences):',kmax
	  write(12,1010)'   avg. size (sequences):',int(float(ksum)/float(knum))
	  write(12,1010)'   max size (bp):',kmaxb
	  write(12,1010)'   avg. size (bp):',int(float(ksumb)/float(knum))
	else
	  write(12,1010)'   number (Contigs):',1
	  write(12,1010)'   max size (sequences):',int(kmax*uscale)
	  write(12,1010)'   avg. size (sequences):',int(ksum*uscale)
	  write(12,1004)'   max size (bp):',euserlength
	  write(12,1004)'   avg. size (bp):',euserlength
	end if
	write(12,1005)' '
	write(12,1005)' IntraMesh Gaps (inside clones)'
	write(12,1000)'   fraction:',float(isum)*scale
	write(12,1010)'   number (internal gaps):',int(inum*uscale)
	write(12,1010)'   max size (bp):',imax
	if(inum .ne. 0)then
	  write(12,1010)'   avg. size: (bp)',isum/inum
	else
	  write(12,1010)'   avg. size: (bp)',0
	end if
	write(12,1005)' '
	write(12,1005)' InterMesh Gaps (between clones)'
	write(12,1000)'   fraction:',float(osum)*scale
	write(12,1010)'   number (external gaps):',int(onum*uscale)
	write(12,1010)'   max size (bp):',omax
	if(onum .ne. 0)then
	  write(12,1010)'   avg. size (bp):',osum/onum
	else
	  write(12,1010)'   avg. size (bp):',0
	end if
	write(12,1010)'   minimum new clones to close:',int(enum*uscale)
	write(12,1005)' '
	write(12,1005)' Regions sequenced more than once, one direction only'
	write(12,1000)'   fraction:',float(vsum)*scale
	write(12,1010)'   number (regions):',int(vnum*uscale)
	write(12,1010)'   max size (bp):',vmax
	if(vnum .ne. 0)then
	  write(12,1010)'   avg. size (bp):',vsum/vnum
	else
	  write(12,1010)'   avg. size (bp):',0
	end if
	write(12,1005)' '
	write(12,1005)' Regions sequenced both directions'
	write(12,1000)'   fraction:',float(bsum)*scale
	write(12,1010)'   number (regions):',int(bnum*uscale)
	write(12,1010)'   max size (bp):',bmax
	if(bnum .ne. 0)then
	  write(12,1010)'   avg. size (bp):',bsum/bnum
	else
	  write(12,1010)'   avg. size (bp):',0
	end if
	write(12,1005)' '
	write(12,1005)' Regions sequenced once only'
	write(12,1000)'   fraction:',float(ssum)*scale
	write(12,1010)'   number (regions):',int(snum*uscale)
	write(12,1010)'   max size (bp):',smax
	if(snum .ne. 0)then
	   write(12,1010)'   avg. size (bp):',ssum/snum
	else
	   write(12,1010)'   avg. size (bp):',0
	end if
	write(12,1005)' '
	write(12,1005)' Sequencing steps required to close'
	write(12,1010)'   IntraMesh gaps (X):',int(isteps*uscale)
	write(12,1010)'   InterMesh gaps (Y):',int(osteps*uscale)
	write(12,1010)'   IntraMesh, one strand -> both strands(Z):',int(ssteps*uscale)
	write(12,1010)'   close all, both strands (Z+2*(X+Y)):'
	1  , int((  ssteps + (isteps + osteps)*2 )* uscale)
	write(12,1005)' '
	write(12,1005)' Total costs'
	sumccost = sumccost * uscale
	sumscost = sumscost * uscale
	write(12,1002)'   Shotgun clone isolation and preparation (A):',sumccost
	write(12,1002)'   Shotgun clone end sequencing (B):',sumscost
	sumpcost =  
	1 int((  ssteps + (isteps + osteps)*2 )* uscale * pcost)
	write(12,1002)'   Custom primer synthesis (C+D+E):',sumpcost
	write(12,1002)'      IntraMesh gaps (C):',isteps*2*uscale*pcost
	write(12,1002)'      InterMesh gaps (D):',osteps*2*uscale*pcost
	write(12,1002)'      IntraMesh 1->2 strands (E):',ssteps*uscale*pcost
	sumicost =  
	1 int((  ssteps + (isteps + osteps)*2 )* uscale * icost)
	write(12,1002)'   Custom primer sequencing (F+G+H): ',sumicost
	write(12,1002)'      IntraMesh gaps (F):',isteps*2*uscale*icost
	write(12,1002)'      InterMesh gaps (G):',osteps*2*uscale*icost
	write(12,1002)'      IntraMesh 1->2 strands (H):',ssteps*uscale*icost
	sumecost =
	1 int( enum * uscale * ecost)
	write(12,1002)'   InterMesh clone isolation (I):',sumecost
	write(12,1005)' '
	write(12,1002)'   Grand Total cost (A+B+C+D+E+F+G+H+I):'
	1    , sumccost + sumscost + sumicost + sumpcost+sumecost
	close(unit=12,dispose='save')
	stop
	end

	subroutine binsort(scratch,toss,nsum,mode)
	implicit none
	integer*4 nsum,mode
c	mode = 0, sort on begin, else, sort on end
	structure /triad/
	   integer*4 begin
	   integer*4 end
	   integer*4 id
	end structure
	record /triad/ scratch(nsum)
	record /triad/ toss(nsum)
	record /triad/ temp
c
	integer*4 ff,fe,fp,ef,ee,ep	!f=front, e=end, p = pointer
	integer*4 out,size,i,j
	logical maintest
	
	if(nsum .lt. 2)return
	size = 1
	do while (size .le. nsum)
	ff = 1
	fe = ff + size - 1

	do while (fe .lt. nsum)
	   ef = fe + 1
	   ee = ef + size - 1
	   if(ee .gt. nsum) ee = nsum
	   fp = ff
	   ep = ef
	   out = ff
	   do while(out .le. ee)
	     if(mode .eq. 0)then
	        maintest = scratch(fp).begin .lt. scratch(ep).begin
	     else
	        maintest = scratch(fp).end .lt. scratch(ep).end
	     end if
	     if(maintest) then
	        toss(out) = scratch(ep)
	        out = out + 1
                if(ep .lt. ee)then
                   ep = ep + 1
                else                !move the rest of the f*  block
                   do while (fp .le. fe)
	              toss(out) = scratch(fp)
	              out = out + 1
                      fp = fp + 1
	           end do
                end if
	     else
	        toss(out) = scratch(fp)
	        out = out + 1
                if(fp .lt. fe)then
                   fp = fp + 1
                else                !move the rest of the e*  block
                   do while (ep .le. ee)
	              toss(out) = scratch(ep)
	              out = out + 1
                      ep = ep + 1
	           end do
                end if
	     end if
	   end do
	   ff = ee + 1
	   fe = ff + size -1
	end do
c
c	Write the toss area back over the input file in
c	preparation for the next step, or from exit from the routine
c	The reason it only goes to "out" is that their may be an "odd"
c	block on the end with no partner to swap with at some iterations.
c
	do out = 1, out - 1
	  scratch(out) = toss(out)
	end do
c
c	set up for the next iteration
c
	size =size + size
	end do
c
c	yes, this is dumb, and the sort should just be rewritten, but
c	the sort is Greatest->least and we need the opposite, so just swap
c
	j= nsum
	do i =1,nsum/2
 	  temp=scratch(i)
	  scratch(i)=scratch(j)
	  scratch(j)=temp
	  j= j - 1
	end do
c
	return
	end
